@@ -1,6 +1,6 @@
 /*
  * The initial "port" of dwm to curses was done by
- * (c) 2007-2008 Marc Andre Tanner <mat at brain-dump dot org>
+ * (c) 2007-2009 Marc Andre Tanner <mat at brain-dump dot org>
  *
  * It is highly inspired by the original X11 dwm and
  * reuses some code of it which is mostly
@@ -25,6 +25,9 @@
 #include <unistd.h>
 #include <stdbool.h>
 #include <errno.h>
+#ifdef __CYGWIN__
+# include <termios.h>
+#endif
 #include "madtty.h"
 
 typedef struct {
@@ -41,6 +44,9 @@ struct Client {
 	uint8_t order;
 	pid_t pid;
 	int pty;
+#ifdef CONFIG_CMDFIFO
+	unsigned short int id;
+#endif
 	short int x;
 	short int y;
 	short int w;
@@ -72,16 +78,24 @@ typedef struct {
 	Action action;
 } Key;
 
-#if defined(HANDLE_MOUSE)
+#ifdef CONFIG_MOUSE
 typedef struct {
 	mmask_t mask;
 	Action action;
 } Button;
-#endif /* HANDLE_MOUSE */
+#endif
 
+#ifdef CONFIG_CMDFIFO
+typedef struct {
+	const char *name;
+	Action action;
+} Cmd;
+#endif
+
+#ifdef CONFIG_STATUSBAR
 enum { BarTop, BarBot, BarOff };
+#endif
 
-#define COLOR(fg, bg) madtty_color_pair(fg, bg)
 #define countof(arr) (sizeof(arr) / sizeof((arr)[0]))
 #define sstrlen(str) (sizeof(str) - 1)
 #define max(x, y) ((x) > (y) ? (x) : (y))
@@ -93,55 +107,75 @@ enum { BarTop, BarBot, BarOff };
 #endif
 
 /* commands for use by keybindings */
-void quit(const char *args[]);
-void create(const char *args[]);
-void startup(const char *args[]);
-void escapekey(const char *args[]);
-void killclient(const char *args[]);
-void focusn(const char *args[]);
-void focusnext(const char *args[]);
-void focusnextnm(const char *args[]);
-void focusprev(const char *args[]);
-void focusprevnm(const char *args[]);
-void toggleminimize(const char *args[]);
-void togglebar(const char *args[]);
-void setmwfact(const char *args[]);
-void setlayout(const char *args[]);
-void redraw(const char *args[]);
-void zoom(const char *args[]);
+static void quit(const char *args[]);
+static void create(const char *args[]);
+static void startup(const char *args[]);
+static void escapekey(const char *args[]);
+static void killclient(const char *args[]);
+static void focusn(const char *args[]);
+static void focusnext(const char *args[]);
+static void focusnextnm(const char *args[]);
+static void focusprev(const char *args[]);
+static void focusprevnm(const char *args[]);
+static void toggleminimize(const char *args[]);
+static void setmwfact(const char *args[]);
+static void setlayout(const char *args[]);
+static void scrollback(const char *args[]);
+static void redraw(const char *args[]);
+static void zoom(const char *args[]);
+static void lock(const char *key[]);
 
-#if defined(HANDLE_MOUSE)
-void mouse_focus(const char *args[]);
-void mouse_fullscreen(const char *args[]);
-void mouse_minimize(const char *args[]);
-void mouse_zoom(const char *args[]);
-#endif /* HANDLE_MOUSE */
+#ifdef CONFIG_STATUSBAR
+enum { ALIGN_LEFT, ALIGN_RIGHT };
+static void togglebar(const char *args[]);
+#endif
 
-void clear_workspace();
-void draw_all(bool border);
-void draw_border(Client *c);
-void drawbar();
-void resize(Client *c, int x, int y, int w, int h);
+#ifdef CONFIG_MOUSE
+static void mouse_focus(const char *args[]);
+static void mouse_fullscreen(const char *args[]);
+static void mouse_minimize(const char *args[]);
+static void mouse_zoom(const char *args[]);
+static void mouse_toggle();
+#endif
 
-unsigned int bh = 1, by, waw, wah, wax, way;
-Client *clients = NULL;
+static void clear_workspace();
+static void draw(Client *c);
+static void draw_all(bool border);
+static void draw_border(Client *c);
+static void resize(Client *c, int x, int y, int w, int h);
+static void eprint(const char *errstr, ...);
+static bool isarrange(void (*func)());
+static void arrange();
+static void focus(Client *c);
+
+static unsigned int waw, wah, wax, way;
+static Client *clients = NULL;
 extern double mwfact;
 
 #include "config.h"
 
-Client *sel = NULL;
-Client *msel = NULL;
+static Client *sel = NULL;
 double mwfact = MWFACT;
-Layout *layout = layouts;
-int statusfd = -1;
-char stext[512];
-int barpos = BARPOS;
-const char *shell;
-bool need_screen_resize = true;
-int width, height;
-bool running = true;
+static Layout *layout = layouts;
+static const char *shell;
+static bool need_screen_resize = true;
+static int width, height, scroll_buf_size = SCROLL_BUF_SIZE;
+static bool running = true;
+static int esc_delay = 100;
 
-void
+#ifdef CONFIG_MOUSE
+# include "mouse.c"
+#endif
+
+#ifdef CONFIG_CMDFIFO
+# include "cmdfifo.c"
+#endif
+
+#ifdef CONFIG_STATUSBAR
+# include "statusbar.c"
+#endif
+
+static void
 eprint(const char *errstr, ...) {
 	va_list ap;
 	va_start(ap, errstr);
@@ -149,7 +183,16 @@ eprint(const char *errstr, ...) {
 	va_end(ap);
 }
 
-void
+static void
+error(const char *errstr, ...) {
+	va_list ap;
+	va_start(ap, errstr);
+	vfprintf(stderr, errstr, ap);
+	va_end(ap);
+	exit(EXIT_FAILURE);
+}
+
+static void
 attach(Client *c) {
 	uint8_t order;
 	if (clients)
@@ -161,7 +204,7 @@ attach(Client *c) {
 		c->order = order;
 }
 
-void
+static void
 attachafter(Client *c, Client *a) { /* attach c after a */
 	uint8_t o;
 	if (c == a)
@@ -180,7 +223,7 @@ attachafter(Client *c, Client *a) { /* attach c after a */
 	}
 }
 
-void
+static void
 detach(Client *c) {
 	Client *d;
 	if (c->prev)
@@ -195,7 +238,7 @@ detach(Client *c) {
 	c->next = c->prev = NULL;
 }
 
-void
+static void
 arrange() {
 	clear_workspace();
 	layout->arrange();
@@ -203,12 +246,12 @@ arrange() {
 	draw_all(true);
 }
 
-bool
+static bool
 isarrange(void (*func)()) {
 	return func == layout->arrange;
 }
 
-void
+static void
 focus(Client *c) {
 	Client *tmp = sel;
 	if (sel == c)
@@ -224,7 +267,7 @@ focus(Client *c) {
 	wrefresh(c->window);
 }
 
-void
+static void
 focusn(const char *args[]) {
 	Client *c;
 
@@ -238,7 +281,7 @@ focusn(const char *args[]) {
 	}
 }
 
-void
+static void
 focusnext(const char *args[]) {
 	Client *c;
 
@@ -252,7 +295,7 @@ focusnext(const char *args[]) {
 		focus(c);
 }
 
-void
+static void
 focusnextnm(const char *args[]) {
 	Client *c;
 
@@ -267,7 +310,7 @@ focusnextnm(const char *args[]) {
 	focus(c);
 }
 
-void
+static void
 focusprev(const char *args[]) {
 	Client *c;
 
@@ -280,7 +323,7 @@ focusprev(const char *args[]) {
 		focus(c);
 }
 
-void
+static void
 focusprevnm(const char *args[]) {
 	Client *c;
 
@@ -295,7 +338,7 @@ focusprevnm(const char *args[]) {
 	focus(c);
 }
 
-void
+static void
 zoom(const char *args[]) {
 	Client *c;
 
@@ -312,7 +355,7 @@ zoom(const char *args[]) {
 	arrange();
 }
 
-void
+static void
 toggleminimize(const char *args[]) {
 	Client *c, *m;
 	unsigned int n;
@@ -352,36 +395,7 @@ toggleminimize(const char *args[]) {
 	arrange();
 }
 
-void
-updatebarpos(void) {
-	by = 0;
-	wax = 0;
-	way = 0;
-	waw = width;
-	wah = height;
-	if (statusfd == -1)
-		return;
-	if (barpos == BarTop) {
-		wah -= bh;
-		way += bh;
-	} else if (barpos == BarBot) {
-		wah -= bh;
-		by = wah;
-	}
-}
-
-void
-togglebar(const char *args[]) {
-	if (barpos == BarOff)
-		barpos = (BARPOS == BarOff) ? BarTop : BARPOS;
-	else
-		barpos = BarOff;
-	updatebarpos();
-	arrange();
-	drawbar();
-}
-
-void
+static void
 setlayout(const char *args[]) {
 	unsigned int i;
 
@@ -399,7 +413,7 @@ setlayout(const char *args[]) {
 	arrange();
 }
 
-void
+static void
 setmwfact(const char *args[]) {
 	double delta;
 
@@ -421,20 +435,35 @@ setmwfact(const char *args[]) {
 	arrange();
 }
 
-void
+static void
+scrollback(const char *args[]) {
+	if (!sel) return;
+
+	if (!args[0] || atoi(args[0]) < 0)
+		madtty_scroll(sel->term, -sel->h/2);
+	else
+		madtty_scroll(sel->term,  sel->h/2);
+
+	draw(sel);
+}
+
+static void
 redraw(const char *args[]) {
 	wrefresh(curscr);
 	draw_all(true);
 }
 
-void
+static void
 draw_border(Client *c) {
 	char *s, t = '\0';
 	int x, y, o;
-	if (sel == c)
-		wattrset(c->window, ATTR_SELECTED);
-	else
-		wattrset(c->window, ATTR_NORMAL);
+	if (sel == c) {
+		wattrset(c->window, SELECTED_ATTR);
+		madtty_color_set(c->window, SELECTED_FG, SELECTED_BG);
+	} else {
+		wattrset(c->window, NORMAL_ATTR);
+		madtty_color_set(c->window, NORMAL_FG, NORMAL_BG);
+	}
 	getyx(c->window, y, x);
 	curs_set(0);
 	mvwhline(c->window, 0, 0, ACS_HLINE, c->w);
@@ -456,7 +485,7 @@ draw_border(Client *c) {
 		curs_set(madtty_cursor(c->term));
 }
 
-void
+static void
 draw_content(Client *c) {
 	if (!c->minimized || isarrange(fullscreen)) {
 		madtty_draw(c->term, c->window, 1, 0);
@@ -465,14 +494,14 @@ draw_content(Client *c) {
 	}
 }
 
-void
+static void
 draw(Client *c) {
 	draw_content(c);
 	draw_border(c);
 	wrefresh(c->window);
 }
 
-void
+static void
 clear_workspace() {
 	unsigned int y;
 	for (y = 0; y < wah; y++)
@@ -480,7 +509,7 @@ clear_workspace() {
 	wnoutrefresh(stdscr);
 }
 
-void
+static void
 draw_all(bool border) {
 	Client *c;
 	curs_set(0);
@@ -506,33 +535,7 @@ draw_all(bool border) {
 	}
 }
 
-void
-drawbar() {
-	int s, l, maxlen = width - 2;
-	char t = stext[maxlen];
-	if (barpos == BarOff || !*stext)
-		return;
-	curs_set(0);
-	attrset(BAR_ATTR);
-	mvaddch(by, 0, '[');
-	stext[maxlen] = '\0';
-	l = strlen(stext);
-	if (BAR_ALIGN_RIGHT)
-		for (s = 0; s + l < maxlen; s++)
-			addch(' ');
-	else
-		for (; l < maxlen; l++)
-			stext[l] = ' ';
-	addstr(stext);
-	stext[maxlen] = t;
-	addch(']');
-	attrset(ATTR_NORMAL);
-	if (sel)
-		curs_set(madtty_cursor(sel->term));
-	refresh();
-}
-
-void
+static void
 escapekey(const char *args[]) {
 	int key;
 	if (sel && (!sel->minimized || isarrange(fullscreen))) {
@@ -545,7 +548,48 @@ escapekey(const char *args[]) {
 	}
 }
 
-void
+/*
+ * Lock the screen until the correct password is entered.
+ * The password can either be specified in config.h which is
+ * not recommended because `strings dvtm` will contain it. If
+ * no password is specified in the configuration file it is read
+ * from the keyboard before the screen is locked.
+ *
+ * NOTE: this function doesn't handle the input from clients. All
+ *       foreground operations are temporarily suspended since the
+ *       function doesn't return.
+ */
+static void
+lock(const char *args[]) {
+	size_t len = 0, i = 0;
+	char buf[16], *pass = buf, c;
+
+	erase();
+	curs_set(0);
+
+	if (args && args[0]) {
+		len = strlen(args[0]);
+		pass = (char *)args[0];
+	} else {
+		mvprintw(LINES / 2, COLS / 2 - 7, "Enter password");
+		while (len < sizeof buf && (c = getch()) != '\n')
+			if (c != ERR)
+				buf[len++] = c;
+	}
+
+	mvprintw(LINES / 2, COLS / 2 - 7, "Screen locked!");
+
+	while (i != len) {
+		for(i = 0; i < len; i++) {
+			if (getch() != pass[i])
+				break;
+		}
+	}
+
+	arrange();
+}
+
+static void
 killclient(const char *args[]) {
 	if (!sel)
 		return;
@@ -553,7 +597,7 @@ killclient(const char *args[]) {
 	kill(-sel->pid, SIGKILL);
 }
 
-int
+static int
 title_escape_seq_handler(madtty_t *term, char *es) {
 	Client *c;
 	unsigned int l;
@@ -569,17 +613,32 @@ title_escape_seq_handler(madtty_t *term, char *es) {
 	return MADTTY_HANDLER_OK;
 }
 
-void
+static void
 create(const char *args[]) {
+	Client *c = calloc(sizeof(Client), 1);
+	if (!c)
+		return;
 	const char *cmd = (args && args[0]) ? args[0] : shell;
 	const char *pargs[] = { "/bin/sh", "-c", cmd, NULL };
-	Client *c = calloc(sizeof(Client), 1);
+#ifdef CONFIG_CMDFIFO
+	c->id = ++client_id;
+	char buf[8];
+	snprintf(buf, sizeof buf, "%d", c->id);
+#endif
+	const char *env[] = {
+		"DVTM", VERSION,
+#ifdef CONFIG_CMDFIFO
+		"DVTM_WINDOW_ID", buf,
+#endif
+		NULL
+	};
+
 	c->window = newwin(wah, waw, way, wax);
-	c->term = madtty_create(height - 1, width);
+	c->term = madtty_create(height - 1, width, scroll_buf_size);
 	c->cmd = cmd;
 	if (args && args[1])
 		strncpy(c->title, args[1], sizeof(c->title));
-	c->pid = madtty_forkpty(c->term, "/bin/sh", pargs, &c->pty);
+	c->pid = madtty_forkpty(c->term, "/bin/sh", pargs, env, &c->pty);
 	madtty_set_data(c->term, c);
 	madtty_set_handler(c->term, title_escape_seq_handler);
 	c->w = width;
@@ -594,7 +653,7 @@ create(const char *args[]) {
 	arrange();
 }
 
-void
+static void
 destroy(Client *c) {
 	if (sel == c)
 		focusnextnm(NULL);
@@ -620,7 +679,7 @@ destroy(Client *c) {
 	arrange();
 }
 
-void
+static void
 move_client(Client *c, int x, int y) {
 	if (c->x == x && c->y == y)
 		return;
@@ -633,7 +692,7 @@ move_client(Client *c, int x, int y) {
 	}
 }
 
-void
+static void
 resize_client(Client *c, int w, int h) {
 	if (c->w == w && c->h == h)
 		return;
@@ -647,13 +706,13 @@ resize_client(Client *c, int w, int h) {
 	madtty_resize(c->term, h - 1, w);
 }
 
-void
+static void
 resize(Client *c, int x, int y, int w, int h) {
 	resize_client(c, w, h);
 	move_client(c, x, y);
 }
 
-bool
+static bool
 is_modifier(unsigned int mod) {
 	unsigned int i;
 	for (i = 0; i < countof(keys); i++) {
@@ -663,7 +722,7 @@ is_modifier(unsigned int mod) {
 	return false;
 }
 
-Key*
+static Key*
 keybinding(unsigned int mod, unsigned int code) {
 	unsigned int i;
 	for (i = 0; i < countof(keys); i++) {
@@ -673,80 +732,7 @@ keybinding(unsigned int mod, unsigned int code) {
 	return NULL;
 }
 
-#if defined(HANDLE_MOUSE)
-
-void
-mouse_focus(const char *args[]) {
-	focus(msel);
-	if (msel->minimized)
-		toggleminimize(NULL);
-}
-
-void
-mouse_fullscreen(const char *args[]) {
-	mouse_focus(NULL);
-	if (isarrange(fullscreen))
-		setlayout(NULL);
-	else
-		setlayout(args);
-}
-
-void
-mouse_minimize(const char *args[]) {
-	focus(msel);
-	toggleminimize(NULL);
-}
-
-void
-mouse_zoom(const char *args[]) {
-	focus(msel);
-	zoom(NULL);
-}
-
-Client*
-get_client_by_coord(int x, int y) {
-	Client *c;
-	if (y < way || y >= wah)
-		return NULL;
-	if (isarrange(fullscreen))
-		return sel;
-	for (c = clients; c; c = c->next) {
-		if (x >= c->x && x < c->x + c->w && y >= c->y && y < c->y + c->h) {
-			debug("mouse event, x: %d y: %d client: %d\n", x, y, c->order);
-			return c;
-		}
-	}
-	return NULL;
-}
-
-void
-handle_mouse() {
-	MEVENT event;
-	unsigned int i;
-	if (getmouse(&event) != OK)
-		return;
-	msel = get_client_by_coord(event.x, event.y);
-	if (!msel)
-		return;
-	for (i = 0; i < countof(buttons); i++)
-		if (event.bstate & buttons[i].mask)
-			buttons[i].action.cmd(buttons[i].action.args);
-	msel = NULL;
-}
-
-void
-mouse_setup() {
-	int i;
-	mmask_t mask;
-	for (i = 0, mask = 0; i < countof(buttons); i++)
-		mask |= buttons[i].mask;
-	if (mask)
-		mousemask(mask, NULL);
-}
-
-#endif /* HANDLE_MOUSE */
-
-Client*
+static Client*
 get_client_by_pid(pid_t pid) {
 	Client *c;
 	for (c = clients; c; c = c->next) {
@@ -756,7 +742,7 @@ get_client_by_pid(pid_t pid) {
 	return NULL;
 }
 
-void
+static void
 sigchld_handler(int sig) {
 	int errsv = errno;
 	int status;
@@ -782,7 +768,7 @@ sigchld_handler(int sig) {
 	errno = errsv;
 }
 
-void
+static void
 sigwinch_handler(int sig) {
 	int errsv = errno;
 
@@ -798,12 +784,12 @@ sigwinch_handler(int sig) {
 	errno = errsv;
 }
 
-void
+static void
 sigterm_handler(int sig) {
 	running = false;
 }
 
-void
+static void
 resize_screen() {
 	debug("resize_screen()\n");
 	if (need_screen_resize) {
@@ -817,20 +803,23 @@ resize_screen() {
 		wrefresh(curscr);
 		refresh();
 	}
+	waw = width;
+	wah = height;
+#ifdef CONFIG_STATUSBAR
 	updatebarpos();
 	drawbar();
+#endif
 	arrange();
 	need_screen_resize = false;
 }
 
-void
+static void
 startup(const char *args[]) {
-	int i;
-	for (i = 0; i < countof(actions); i++)
+	for (int i = 0; i < countof(actions); i++)
 		actions[i].cmd(actions[i].args);
 }
 
-void
+static void
 setup() {
 	if (!(shell = getenv("SHELL")))
 		shell = "/bin/sh";
@@ -839,7 +828,7 @@ setup() {
 	start_color();
 	noecho();
 	keypad(stdscr, TRUE);
-#if defined(HANDLE_MOUSE)
+#ifdef CONFIG_MOUSE
 	mouse_setup();
 #endif
 	raw();
@@ -850,29 +839,62 @@ setup() {
 	signal(SIGWINCH, sigwinch_handler);
 	signal(SIGCHLD, sigchld_handler);
 	signal(SIGTERM, sigterm_handler);
+	set_escdelay(esc_delay);
 }
 
-void
+static void
 cleanup() {
 	endwin();
+#ifdef CONFIG_STATUSBAR
 	if (statusfd > 0)
 		close(statusfd);
+#endif
+#ifdef CONFIG_CMDFIFO
+	if (cmdfd > 0)
+		close(cmdfd);
+	if (cmdpath)
+		unlink(cmdpath);
+#endif
 }
 
-void
+static void
 quit(const char *args[]) {
 	cleanup();
 	exit(EXIT_SUCCESS);
 }
 
-void
+static void
 usage() {
 	cleanup();
-	eprint("usage: dvtm [-v] [-m mod] [-s status] [cmd...]\n");
+	eprint("usage: dvtm [-v] [-m mod] [-d escdelay] [-h n] "
+#ifdef CONFIG_STATUSBAR
+		"[-s status-fifo] "
+#endif
+#ifdef CONFIG_CMDFIFO
+		"[-c cmd-fifo] "
+#endif
+		"[cmd...]\n");
 	exit(EXIT_FAILURE);
 }
 
-bool
+static int
+open_or_create_fifo(const char *name) {
+	struct stat info;
+	int fd;
+open:
+	if ((fd = open(name, O_RDWR|O_NONBLOCK)) == -1) {
+		if (errno == ENOENT && !mkfifo(name, S_IRUSR|S_IWUSR))
+			goto open;
+		error("%s\n", strerror(errno));
+	}
+	if (fstat(fd, &info) == -1)
+		error("%s\n", strerror(errno));
+	if (!S_ISFIFO(info.st_mode))
+		error("%s is not a named pipe\n", name);
+	return fd;
+}
+
+static bool
 parse_args(int argc, char *argv[]) {
 	int arg;
 	bool init = false;
@@ -886,35 +908,44 @@ parse_args(int argc, char *argv[]) {
 			create(args);
 			continue;
 		}
+		if (argv[arg][1] != 'v' && (arg + 1) >= argc)
+			usage();
 		switch (argv[arg][1]) {
 			case 'v':
-				puts("dvtm-"VERSION" (c) 2007-2008 Marc Andre Tanner");
+				puts("dvtm-"VERSION" (c) 2007-2009 Marc Andre Tanner");
 				exit(EXIT_SUCCESS);
-			case 'm':
-				if (++arg >= argc)
-					usage();
-				unsigned int i;
-				char *mod = argv[arg];
+			case 'm': {
+				char *mod = argv[++arg];
 				if (mod[0] == '^' && mod[1])
 					*mod = CTRL(mod[1]);
-				for (i = 0; i < countof(keys); i++)
+				for (int i = 0; i < countof(keys); i++)
 					keys[i].mod = *mod;
 				break;
+			}
+			case 'd':
+				esc_delay = atoi(argv[++arg]);
+				if (esc_delay < 50)
+					esc_delay = 50;
+				else if (esc_delay > 1000)
+					esc_delay = 1000;
+				break;
+			case 'h':
+				scroll_buf_size = atoi(argv[++arg]);
+				break;
+#ifdef CONFIG_STATUSBAR
 			case 's':
-				if (++arg >= argc)
-					usage();
-				struct stat info;
-				if ((statusfd = open(argv[arg], O_RDONLY|O_NONBLOCK)) == -1 ||
-				    fstat(statusfd, &info) == -1) {
-					perror("status");
-					exit(EXIT_FAILURE);
-				}
-				if (!S_ISFIFO(info.st_mode)) {
-					eprint("%s is not a named pipe.\n", argv[arg]);
-					exit(EXIT_FAILURE);
-				}
+				statusfd = open_or_create_fifo(argv[++arg]);
 				updatebarpos();
 				break;
+#endif
+#ifdef CONFIG_CMDFIFO
+			case 'c':
+				cmdfd = open_or_create_fifo(argv[++arg]);
+				if (!(cmdpath = get_realpath(argv[arg])))
+					error("%s\n", strerror(errno));
+				setenv("DVTM_CMD_FIFO", cmdpath, 1);
+				break;
+#endif
 			default:
 				usage();
 		}
@@ -928,6 +959,7 @@ main(int argc, char *argv[]) {
 		setup();
 		startup(NULL);
 	}
+
 	while (running) {
 		Client *c, *t;
 		int r, nfds = 0;
@@ -939,11 +971,18 @@ main(int argc, char *argv[]) {
 		FD_ZERO(&rd);
 		FD_SET(STDIN_FILENO, &rd);
 
+#ifdef CONFIG_CMDFIFO
+		if (cmdfd != -1) {
+			FD_SET(cmdfd, &rd);
+			nfds = cmdfd;
+		}
+#endif
+#ifdef CONFIG_STATUSBAR
 		if (statusfd != -1) {
 			FD_SET(statusfd, &rd);
 			nfds = max(nfds, statusfd);
 		}
-
+#endif
 		for (c = clients; c; ) {
 			if (c->died) {
 				t = c->next;
@@ -969,11 +1008,11 @@ main(int argc, char *argv[]) {
 			int code = getch();
 			Key *key;
 			if (code >= 0) {
-#if defined(HANDLE_MOUSE)
+#ifdef CONFIG_MOUSE
 				if (code == KEY_MOUSE) {
 					handle_mouse();
 				} else
-#endif /* HANDLE_MOUSE */
+#endif /* CONFIG_MOUSE */
 				if (is_modifier(code)) {
 					int mod = code;
 					code = getch();
@@ -1011,26 +1050,14 @@ main(int argc, char *argv[]) {
 				continue;
 		}
 
-		if (statusfd != -1 && FD_ISSET(statusfd, &rd)) {
-			char *p;
-			switch (r = read(statusfd, stext, sizeof stext - 1)) {
-				case -1:
-					strncpy(stext, strerror(errno), sizeof stext - 1);
-					stext[sizeof stext - 1] = '\0';
-					statusfd = -1;
-					break;
-				case 0:
-					statusfd = -1;
-					break;
-				default:
-					stext[r] = '\0'; p = stext + strlen(stext) - 1;
-					for (; p >= stext && *p == '\n'; *p-- = '\0');
-					for (; p >= stext && *p != '\n'; --p);
-					if (p > stext)
-						strncpy(stext, p + 1, sizeof stext);
-					drawbar();
-			}
-		}
+#ifdef CONFIG_CMDFIFO
+		if (cmdfd != -1 && FD_ISSET(cmdfd, &rd))
+			handle_cmdfifo();
+#endif
+#ifdef CONFIG_STATUSBAR
+		if (statusfd != -1 && FD_ISSET(statusfd, &rd))
+			handle_statusbar();
+#endif
 
 		for (c = clients; c; ) {
 			if (FD_ISSET(c->pty, &rd)) {
